@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1996-2019. All Rights Reserved.
+%% Copyright Ericsson AB 1996-2023. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -26,7 +26,8 @@
 	 start_child/2, restart_child/2,
 	 delete_child/2, terminate_child/2,
 	 which_children/1, count_children/1,
-	 check_childspecs/1, get_childspec/2]).
+	 check_childspecs/1, check_childspecs/2,
+	 get_childspec/2]).
 
 %% Internal exports
 -export([init/1, handle_call/3, handle_cast/2, handle_info/2,
@@ -61,40 +62,44 @@
 
 %%--------------------------------------------------------------------------
 
--type child()    :: 'undefined' | pid().
--type child_id() :: term().
--type mfargs()   :: {M :: module(), F :: atom(), A :: [term()] | undefined}.
--type modules()  :: [module()] | 'dynamic'.
--type restart()  :: 'permanent' | 'transient' | 'temporary'.
--type shutdown() :: 'brutal_kill' | timeout().
--type worker()   :: 'worker' | 'supervisor'.
--type sup_name() :: {'local', Name :: atom()}
-                  | {'global', Name :: atom()}
-                  | {'via', Module :: module(), Name :: any()}.
--type sup_ref()  :: (Name :: atom())
-                  | {Name :: atom(), Node :: node()}
-                  | {'global', Name :: atom()}
-                  | {'via', Module :: module(), Name :: any()}
-                  | pid().
--type child_spec() :: #{id := child_id(),       % mandatory
-			start := mfargs(),      % mandatory
-			restart => restart(),   % optional
-			shutdown => shutdown(), % optional
-			type => worker(),       % optional
-			modules => modules()}   % optional
-                    | {Id :: child_id(),
-                       StartFunc :: mfargs(),
-                       Restart :: restart(),
-                       Shutdown :: shutdown(),
-                       Type :: worker(),
-                       Modules :: modules()}.
+-type auto_shutdown() :: 'never' | 'any_significant' | 'all_significant'.
+-type child()         :: 'undefined' | pid().
+-type child_id()      :: term().
+-type mfargs()        :: {M :: module(), F :: atom(), A :: [term()] | undefined}.
+-type modules()       :: [module()] | 'dynamic'.
+-type restart()       :: 'permanent' | 'transient' | 'temporary'.
+-type significant()   :: boolean().
+-type shutdown()      :: 'brutal_kill' | timeout().
+-type worker()        :: 'worker' | 'supervisor'.
+-type sup_name()      :: {'local', Name :: atom()}
+                       | {'global', Name :: term()}
+                       | {'via', Module :: module(), Name :: any()}.
+-type sup_ref()       :: (Name :: atom())
+                       | {Name :: atom(), Node :: node()}
+                       | {'global', Name :: term()}
+                       | {'via', Module :: module(), Name :: any()}
+                       | pid().
+-type child_spec()    :: #{id := child_id(),             % mandatory
+			   start := mfargs(),            % mandatory
+			   restart => restart(),         % optional
+			   significant => significant(), % optional
+			   shutdown => shutdown(),       % optional
+			   type => worker(),             % optional
+			   modules => modules()}         % optional
+                       | {Id :: child_id(),
+                          StartFunc :: mfargs(),
+                          Restart :: restart(),
+                          Shutdown :: shutdown(),
+                          Type :: worker(),
+                          Modules :: modules()}.
 
 -type strategy() :: 'one_for_all' | 'one_for_one'
                   | 'rest_for_one' | 'simple_one_for_one'.
 
--type sup_flags() :: #{strategy => strategy(),         % optional
-		       intensity => non_neg_integer(), % optional
-		       period => pos_integer()}        % optional
+-type sup_flags() :: #{strategy => strategy(),           % optional
+		       intensity => non_neg_integer(),   % optional
+		       period => pos_integer(),          % optional
+		       auto_shutdown => auto_shutdown()} % optional
                    | {RestartStrategy :: strategy(),
                       Intensity :: non_neg_integer(),
                       Period :: pos_integer()}.
@@ -102,9 +107,10 @@
 
 %%--------------------------------------------------------------------------
 %% Defaults
--define(default_flags, #{strategy  => one_for_one,
-			 intensity => 1,
-			 period    => 5}).
+-define(default_flags, #{strategy      => one_for_one,
+			 intensity     => 1,
+			 period        => 5,
+			 auto_shutdown => never}).
 -define(default_child_spec, #{restart  => permanent,
 			      type     => worker}).
 %% Default 'shutdown' is 5000 for workers and infinity for supervisors.
@@ -119,21 +125,23 @@
 		id              :: child_id(),
 		mfargs          :: mfargs(),
 		restart_type    :: restart(),
+		significant     :: significant(),
 		shutdown        :: shutdown(),
 		child_type      :: worker(),
 		modules = []    :: modules()}).
 -type child_rec() :: #child{}.
 
 -record(state, {name,
-		strategy               :: strategy() | 'undefined',
+		strategy = one_for_one:: strategy(),
 		children = {[],#{}}    :: children(), % Ids in start order
                 dynamics               :: {'maps', #{pid() => list()}}
                                         | {'mapsets', #{pid() => []}}
                                         | 'undefined',
-		intensity              :: non_neg_integer() | 'undefined',
-		period                 :: pos_integer() | 'undefined',
+		intensity = 1          :: non_neg_integer(),
+		period    = 5          :: pos_integer(),
 		restarts = [],
 		dynamic_restarts = 0   :: non_neg_integer(),
+		auto_shutdown = never  :: auto_shutdown(),
 	        module,
 	        args}).
 -type state() :: #state{}.
@@ -142,6 +150,7 @@
 -define(is_temporary(_Child_), _Child_#child.restart_type=:=temporary).
 -define(is_transient(_Child_), _Child_#child.restart_type=:=transient).
 -define(is_permanent(_Child_), _Child_#child.restart_type=:=permanent).
+-define(is_significant(_Child_), _Child_#child.significant=:=true).
 
 -callback init(Args :: term()) ->
     {ok, {SupFlags :: sup_flags(), [ChildSpec :: child_spec()]}}
@@ -152,7 +161,7 @@
 %%% ---------------------------------------------------
 %%% This is a general process supervisor built upon gen_server.erl.
 %%% Servers/processes should/could also be built using gen_server.erl.
-%%% SupName = {local, atom()} | {global, atom()}.
+%%% SupName = {local, atom()} | {global, term()}.
 %%% ---------------------------------------------------
 
 -type startlink_err() :: {'already_started', pid()}
@@ -246,7 +255,7 @@ which_children(Supervisor) ->
       Count :: {specs, ChildSpecCount :: non_neg_integer()}
              | {active, ActiveProcessCount :: non_neg_integer()}
              | {supervisors, ChildSupervisorCount :: non_neg_integer()}
-             |{workers, ChildWorkerCount :: non_neg_integer()}.
+             | {workers, ChildWorkerCount :: non_neg_integer()}.
 count_children(Supervisor) ->
     call(Supervisor, count_children).
 
@@ -256,12 +265,32 @@ call(Supervisor, Req) ->
 -spec check_childspecs(ChildSpecs) -> Result when
       ChildSpecs :: [child_spec()],
       Result :: 'ok' | {'error', Error :: term()}.
-check_childspecs(ChildSpecs) when is_list(ChildSpecs) ->
-    case check_startspec(ChildSpecs) of
+check_childspecs(ChildSpecs) ->
+    check_childspecs(ChildSpecs, undefined).
+
+-spec check_childspecs(ChildSpecs, AutoShutdown) -> Result when
+      ChildSpecs :: [child_spec()],
+      AutoShutdown :: undefined | auto_shutdown(),
+      Result :: 'ok' | {'error', Error :: term()}.
+check_childspecs(ChildSpecs, AutoShutdown) when is_list(ChildSpecs) ->
+    check_childspecs1(ChildSpecs, AutoShutdown);
+check_childspecs(X, _AutoShutdown) -> {error, {badarg, X}}.
+
+check_childspecs1(ChildSpecs, undefined) ->
+    check_childspecs2(ChildSpecs, undefined);
+check_childspecs1(ChildSpecs, never) ->
+    check_childspecs2(ChildSpecs, never);
+check_childspecs1(ChildSpecs, any_significant) ->
+    check_childspecs2(ChildSpecs, any_significant);
+check_childspecs1(ChildSpecs, all_significant) ->
+    check_childspecs2(ChildSpecs, all_significant);
+check_childspecs1(_, X) -> {error, {badarg, X}}.
+
+check_childspecs2(ChildSpecs, AutoShutdown) ->
+    case check_startspec(ChildSpecs, AutoShutdown) of
 	{ok, _} -> ok;
 	Error -> {error, Error}
-    end;
-check_childspecs(X) -> {error, {badarg, X}}.
+    end.
 
 %%%-----------------------------------------------------------------
 %%% Called by release_handler during upgrade
@@ -316,7 +345,7 @@ init({SupName, Mod, Args}) ->
 
 init_children(State, StartSpec) ->
     SupName = State#state.name,
-    case check_startspec(StartSpec) of
+    case check_startspec(StartSpec, State#state.auto_shutdown) of
         {ok, Children} ->
             case start_children(Children, SupName) of
                 {ok, NChildren} ->
@@ -330,7 +359,7 @@ init_children(State, StartSpec) ->
     end.
 
 init_dynamic(State, [StartSpec]) ->
-    case check_startspec([StartSpec]) of
+    case check_startspec([StartSpec], State#state.auto_shutdown) of
         {ok, Children} ->
 	    {ok, dyn_init(State#state{children = Children})};
         Error ->
@@ -342,7 +371,7 @@ init_dynamic(_State, StartSpec) ->
 %%-----------------------------------------------------------------
 %% Func: start_children/2
 %% Args: Children = children() % Ids in start order
-%%       SupName = {local, atom()} | {global, atom()} | {pid(), Mod}
+%%       SupName = {local, atom()} | {global, term()} | {pid(), Mod}
 %% Purpose: Start all children.  The new map contains #child's
 %%          with pids.
 %% Returns: {ok, NChildren} | {error, NChildren, Reason}
@@ -421,7 +450,7 @@ handle_call({start_child, EArgs}, _From, State) when ?is_simple(State) ->
     end;
 
 handle_call({start_child, ChildSpec}, _From, State) ->
-    case check_childspec(ChildSpec) of
+    case check_childspec(ChildSpec, State#state.auto_shutdown) of
 	{ok, Child} ->
 	    {Resp, NState} = handle_start_child(Child, State),
 	    {reply, Resp, NState};
@@ -630,14 +659,14 @@ code_change(_, State, _) ->
     end.
 
 update_childspec(State, StartSpec) when ?is_simple(State) ->
-    case check_startspec(StartSpec) of
+    case check_startspec(StartSpec, State#state.auto_shutdown) of
         {ok, {[_],_}=Children} ->
             {ok, State#state{children = Children}};
         Error ->
             {error, Error}
     end;
 update_childspec(State, StartSpec) ->
-    case check_startspec(StartSpec) of
+    case check_startspec(StartSpec, State#state.auto_shutdown) of
 	{ok, Children} ->
 	    OldC = State#state.children, % In reverse start order !
 	    NewC = update_childspec1(OldC, Children, []),
@@ -708,20 +737,54 @@ do_restart(Reason, Child, State) when ?is_permanent(Child) ->
     restart(Child, State);
 do_restart(normal, Child, State) ->
     NState = del_child(Child, State),
-    {ok, NState};
+    do_auto_shutdown(Child, NState);
 do_restart(shutdown, Child, State) ->
     NState = del_child(Child, State),
-    {ok, NState};
+    do_auto_shutdown(Child, NState);
 do_restart({shutdown, _Term}, Child, State) ->
     NState = del_child(Child, State),
-    {ok, NState};
+    do_auto_shutdown(Child, NState);
 do_restart(Reason, Child, State) when ?is_transient(Child) ->
     ?report_error(child_terminated, Reason, Child, State#state.name),
     restart(Child, State);
 do_restart(Reason, Child, State) when ?is_temporary(Child) ->
     ?report_error(child_terminated, Reason, Child, State#state.name),
     NState = del_child(Child, State),
-    {ok, NState}.
+    do_auto_shutdown(Child, NState).
+
+do_auto_shutdown(_Child, State=#state{auto_shutdown = never}) ->
+    {ok, State};
+do_auto_shutdown(Child, State) when not ?is_significant(Child) ->
+    {ok, State};
+do_auto_shutdown(_Child, State=#state{auto_shutdown = any_significant}) ->
+    {shutdown, State};
+do_auto_shutdown(_Child, State=#state{auto_shutdown = all_significant})
+  when ?is_simple(State) ->
+    case dyn_size(State) of
+	0 ->
+	    {shutdown, State};
+	_ ->
+	    {ok, State}
+    end;
+do_auto_shutdown(_Child, State=#state{auto_shutdown = all_significant}) ->
+    case
+	children_any(
+	    fun
+		(_, #child{pid = undefined}) ->
+		    false;
+		(_, #child{significant = true}) ->
+		    true;
+		(_, _) ->
+		    false
+	    end,
+	    State#state.children
+	)
+    of
+	true ->
+	    {ok, State};
+	false ->
+	    {shutdown, State}
+    end.
 
 restart(Child, State) ->
     case add_restart(State) of
@@ -729,7 +792,7 @@ restart(Child, State) ->
 	    case restart(NState#state.strategy, Child, NState) of
 		{{try_again, TryAgainId}, NState2} ->
 		    %% Leaving control back to gen_server before
-		    %% trying again. This way other incoming requsts
+		    %% trying again. This way other incoming requests
 		    %% for the supervisor can be handled - e.g. a
 		    %% shutdown request for the supervisor or the
 		    %% child.
@@ -816,7 +879,7 @@ try_again_restart(TryAgainId) ->
 %%-----------------------------------------------------------------
 %% Func: terminate_children/2
 %% Args: Children = children() % Ids in termination order
-%%       SupName = {local, atom()} | {global, atom()} | {pid(),Mod}
+%%       SupName = {local, atom()} | {global, term()} | {pid(),Mod}
 %% Returns: NChildren = children() % Ids in startup order
 %%                                 % (reversed termination order)
 %%-----------------------------------------------------------------
@@ -835,11 +898,9 @@ terminate_children(Children, SupName) ->
     NChildren.
 
 do_terminate(Child, SupName) when is_pid(Child#child.pid) ->
-    case shutdown(Child#child.pid, Child#child.shutdown) of
+    case shutdown(Child) of
         ok ->
             ok;
-        {error, normal} when not (?is_permanent(Child)) ->
-            ok;
         {error, OtherReason} ->
             ?report_error(shutdown_error, OtherReason, Child, SupName)
     end,
@@ -857,65 +918,65 @@ do_terminate(_Child, _SupName) ->
 %% supervisor. 
 %% Returns: ok | {error, OtherReason}  (this should be reported)
 %%-----------------------------------------------------------------
-shutdown(Pid, brutal_kill) ->
-    case monitor_child(Pid) of
-	ok ->
-	    exit(Pid, kill),
-	    receive
-		{'DOWN', _MRef, process, Pid, killed} ->
-		    ok;
-		{'DOWN', _MRef, process, Pid, OtherReason} ->
-		    {error, OtherReason}
-	    end;
-	{error, Reason} ->      
-	    {error, Reason}
+shutdown(#child{pid=Pid, shutdown=brutal_kill} = Child) ->
+    Mon = monitor(process, Pid),
+    exit(Pid, kill),
+    receive
+        {'DOWN', Mon, process, Pid, Reason0} ->
+            case unlink_flush(Pid, Reason0) of
+                killed ->
+                    ok;
+                shutdown when not (?is_permanent(Child)) ->
+                    ok;
+                {shutdown, _} when not (?is_permanent(Child)) ->
+                    ok;
+                normal when not (?is_permanent(Child)) ->
+                    ok;
+                Reason ->
+                    {error, Reason}
+            end
     end;
-shutdown(Pid, Time) ->
-    case monitor_child(Pid) of
-	ok ->
-	    exit(Pid, shutdown), %% Try to shutdown gracefully
-	    receive 
-		{'DOWN', _MRef, process, Pid, shutdown} ->
-		    ok;
-		{'DOWN', _MRef, process, Pid, OtherReason} ->
-		    {error, OtherReason}
-	    after Time ->
-		    exit(Pid, kill),  %% Force termination.
-		    receive
-			{'DOWN', _MRef, process, Pid, OtherReason} ->
-			    {error, OtherReason}
-		    end
-	    end;
-	{error, Reason} ->      
-	    {error, Reason}
+shutdown(#child{pid=Pid, shutdown=Time} = Child) ->
+    Mon = monitor(process, Pid),
+    exit(Pid, shutdown),
+    receive
+        {'DOWN', Mon, process, Pid, Reason0} ->
+            case unlink_flush(Pid, Reason0) of
+                shutdown ->
+                    ok;
+                {shutdown, _} when not (?is_permanent(Child)) ->
+                    ok;
+                normal when not (?is_permanent(Child)) ->
+                    ok;
+                Reason ->
+                    {error, Reason}
+            end
+    after Time ->
+        exit(Pid, kill),
+        receive
+            {'DOWN', Mon, process, Pid, Reason0} ->
+                case unlink_flush(Pid, Reason0) of
+                    shutdown ->
+                        ok;
+                    {shutdown, _} when not (?is_permanent(Child)) ->
+                        ok;
+                    normal when not (?is_permanent(Child)) ->
+                        ok;
+                    Reason ->
+                        {error, Reason}
+                end
+        end
     end.
 
-%% Help function to shutdown/2 switches from link to monitor approach
-monitor_child(Pid) ->
-    
-    %% Do the monitor operation first so that if the child dies 
-    %% before the monitoring is done causing a 'DOWN'-message with
-    %% reason noproc, we will get the real reason in the 'EXIT'-message
-    %% unless a naughty child has already done unlink...
-    erlang:monitor(process, Pid),
+unlink_flush(Pid, DefaultReason) ->
+    %% We call unlink in order to guarantee that the 'EXIT' has arrived
+    %% from the dead process. See the unlink docs for details.
     unlink(Pid),
-
     receive
-	%% If the child dies before the unlik we must empty
-	%% the mail-box of the 'EXIT'-message and the 'DOWN'-message.
-	{'EXIT', Pid, Reason} -> 
-	    receive 
-		{'DOWN', _, process, Pid, _} ->
-		    {error, Reason}
-	    end
-    after 0 -> 
-	    %% If a naughty child did unlink and the child dies before
-	    %% monitor the result will be that shutdown/2 receives a 
-	    %% 'DOWN'-message with reason noproc.
-	    %% If the child should die after the unlink there
-	    %% will be a 'DOWN'-message with a correct reason
-	    %% that will be handled in shutdown/2. 
-	    ok   
+        {'EXIT', Pid, Reason} ->
+            Reason
+    after 0 ->
+        DefaultReason
     end.
 
 %%-----------------------------------------------------------------
@@ -929,39 +990,36 @@ monitor_child(Pid) ->
 %%-----------------------------------------------------------------
 terminate_dynamic_children(State) ->
     Child = get_dynamic_child(State),
-    {Pids, EStack0} = monitor_dynamic_children(Child,State),
+    Pids = dyn_fold(
+        fun
+            (P, Acc) when is_pid(P) ->
+                Mon = monitor(process, P),
+                case Child#child.shutdown of
+                    brutal_kill -> exit(P, kill);
+                    _ -> exit(P, shutdown)
+                end,
+                Acc#{{P, Mon} => true};
+            (?restarting(_), Acc) ->
+                Acc
+        end,
+        #{},
+        State
+    ),
+    TRef = case Child#child.shutdown of
+        brutal_kill ->
+            undefined;
+        infinity ->
+            undefined;
+        Time ->
+            erlang:start_timer(Time, self(), kill)
+    end,
     Sz = maps:size(Pids),
-    EStack = case Child#child.shutdown of
-                 brutal_kill ->
-                     maps:fold(fun(P, _, _) -> exit(P, kill) end, ok, Pids),
-                     wait_dynamic_children(Child, Pids, Sz, undefined, EStack0);
-                 infinity ->
-                     maps:fold(fun(P, _, _) -> exit(P, shutdown) end, ok, Pids),
-                     wait_dynamic_children(Child, Pids, Sz, undefined, EStack0);
-                 Time ->
-                     maps:fold(fun(P, _, _) -> exit(P, shutdown) end, ok, Pids),
-                     TRef = erlang:start_timer(Time, self(), kill),
-                     wait_dynamic_children(Child, Pids, Sz, TRef, EStack0)
-             end,
+    EStack = wait_dynamic_children(Child, Pids, Sz, TRef, #{}),
     %% Unroll stacked errors and report them
-    maps:fold(fun(Reason, Ls, _) ->
+    maps:foreach(fun(Reason, Ls) ->
                       ?report_error(shutdown_error, Reason,
                                    Child#child{pid=Ls}, State#state.name)
-              end, ok, EStack).
-
-monitor_dynamic_children(Child,State) ->
-    dyn_fold(fun(P,{Pids, EStack}) when is_pid(P) ->
-                     case monitor_child(P) of
-                         ok ->
-                             {maps:put(P, P, Pids), EStack};
-                         {error, normal} when not (?is_permanent(Child)) ->
-                             {Pids, EStack};
-                         {error, Reason} ->
-                             {Pids, maps_prepend(Reason, P, EStack)}
-                     end;
-                (?restarting(_), {Pids, EStack}) ->
-                     {Pids, EStack}
-             end, {maps:new(), maps:new()}, State).
+              end, EStack).
 
 wait_dynamic_children(_Child, _Pids, 0, undefined, EStack) ->
     EStack;
@@ -978,34 +1036,55 @@ wait_dynamic_children(_Child, _Pids, 0, TRef, EStack) ->
 wait_dynamic_children(#child{shutdown=brutal_kill} = Child, Pids, Sz,
                       TRef, EStack) ->
     receive
-        {'DOWN', _MRef, process, Pid, killed} ->
-            wait_dynamic_children(Child, maps:remove(Pid, Pids), Sz-1,
-                                  TRef, EStack);
+        {'DOWN', Mon, process, Pid, Reason0}
+          when is_map_key({Pid, Mon}, Pids) ->
+            case unlink_flush(Pid, Reason0) of
+                killed ->
+                    wait_dynamic_children(Child, maps:remove({Pid, Mon}, Pids),
+                                          Sz-1, TRef, EStack);
 
-        {'DOWN', _MRef, process, Pid, Reason} ->
-            wait_dynamic_children(Child, maps:remove(Pid, Pids), Sz-1,
-                                  TRef, maps_prepend(Reason, Pid, EStack))
+                shutdown when not (?is_permanent(Child)) ->
+                    wait_dynamic_children(Child, maps:remove({Pid, Mon}, Pids),
+                                          Sz-1, TRef, EStack);
+
+                {shutdown, _} when not (?is_permanent(Child)) ->
+                    wait_dynamic_children(Child, maps:remove({Pid, Mon}, Pids),
+                                          Sz-1, TRef, EStack);
+
+                normal when not (?is_permanent(Child)) ->
+                    wait_dynamic_children(Child, maps:remove({Pid, Mon}, Pids),
+                                          Sz-1, TRef, EStack);
+                Reason ->
+                    wait_dynamic_children(Child, maps:remove({Pid, Mon}, Pids),
+                                          Sz-1, TRef, maps_prepend(Reason, Pid,
+                                                                   EStack))
+            end
     end;
 wait_dynamic_children(Child, Pids, Sz, TRef, EStack) ->
     receive
-        {'DOWN', _MRef, process, Pid, shutdown} ->
-            wait_dynamic_children(Child, maps:remove(Pid, Pids), Sz-1,
-                                  TRef, EStack);
+        {'DOWN', Mon, process, Pid, Reason0}
+          when is_map_key({Pid, Mon}, Pids) ->
+            case unlink_flush(Pid, Reason0) of
+                shutdown ->
+                    wait_dynamic_children(Child, maps:remove({Pid, Mon}, Pids),
+                                          Sz-1, TRef, EStack);
 
-        {'DOWN', _MRef, process, Pid, {shutdown, _}} ->
-            wait_dynamic_children(Child, maps:remove(Pid, Pids), Sz-1,
-                                  TRef, EStack);
+                {shutdown, _} when not (?is_permanent(Child)) ->
+                    wait_dynamic_children(Child, maps:remove({Pid, Mon}, Pids),
+                                          Sz-1, TRef, EStack);
 
-        {'DOWN', _MRef, process, Pid, normal} when not (?is_permanent(Child)) ->
-            wait_dynamic_children(Child, maps:remove(Pid, Pids), Sz-1,
-                                  TRef, EStack);
+                normal when not (?is_permanent(Child)) ->
+                    wait_dynamic_children(Child, maps:remove({Pid, Mon}, Pids),
+                                          Sz-1, TRef, EStack);
 
-        {'DOWN', _MRef, process, Pid, Reason} ->
-            wait_dynamic_children(Child, maps:remove(Pid, Pids), Sz-1,
-                                  TRef, maps_prepend(Reason, Pid, EStack));
+                Reason ->
+                    wait_dynamic_children(Child, maps:remove({Pid, Mon}, Pids),
+                                          Sz-1, TRef, maps_prepend(Reason, Pid,
+                                                                   EStack))
+            end;
 
         {timeout, TRef, kill} ->
-            maps:fold(fun(P, _, _) -> exit(P, kill) end, ok, Pids),
+            maps:foreach(fun({P, _}, _) -> exit(P, kill) end, Pids),
             wait_dynamic_children(Child, Pids, Sz, undefined, EStack)
     end.
 
@@ -1118,7 +1197,7 @@ find_dynamic_child(Pid, State) ->
             error
     end.
 
-%% Given the pid, find the child record for a non-dyanamic child.
+%% Given the pid, find the child record for a non-dynamic child.
 -spec find_child_by_pid(IdOrPid, state()) -> {ok,child_rec()} | error when
       IdOrPid :: pid() | {restarting,pid()}.
 find_child_by_pid(Pid,#state{children={_Ids,Db}}) ->
@@ -1201,13 +1280,23 @@ children_to_list(_Fun,[],_Db,Acc) ->
 children_fold(Fun,Init,{_Ids,Db}) ->
     maps:fold(Fun, Init, Db).
 
+%% The order is not important - so ignore Ids
+children_any(Pred, {_Ids, Db}) ->
+    Iter=maps:iterator(Db),
+    children_any1(Pred, maps:next(Iter)).
+
+children_any1(_Pred, none) ->
+    false;
+children_any1(Pred, {Key, Value, Iter}) ->
+    Pred(Key, Value) orelse children_any1(Pred, maps:next(Iter)).
+
 -spec append(children(), children()) -> children().
 append({Ids1,Db1},{Ids2,Db2}) ->
     {Ids1++Ids2,maps:merge(Db1,Db2)}.
 
 %%-----------------------------------------------------------------
 %% Func: init_state/4
-%% Args: SupName = {local, atom()} | {global, atom()} | self
+%% Args: SupName = {local, atom()} | {global, term()} | self
 %%       Type = {Strategy, MaxIntensity, Period}
 %%         Strategy = one_for_one | one_for_all | simple_one_for_one |
 %%                    rest_for_one
@@ -1221,14 +1310,17 @@ append({Ids1,Db1},{Ids2,Db2}) ->
 init_state(SupName, Type, Mod, Args) ->
     set_flags(Type, #state{name = supname(SupName,Mod),
 			   module = Mod,
-			   args = Args}).
+			   args = Args,
+			   auto_shutdown = never}).
 
 set_flags(Flags, State) ->
     try check_flags(Flags) of
-	#{strategy := Strategy, intensity := MaxIntensity, period := Period} ->
+	#{strategy := Strategy, intensity := MaxIntensity, period := Period,
+	  auto_shutdown := AutoShutdown} ->
 	    {ok, State#state{strategy = Strategy,
 			     intensity = MaxIntensity,
-			     period = Period}}
+			     period = Period,
+			     auto_shutdown = AutoShutdown}}
     catch
 	Thrown -> Thrown
     end.
@@ -1238,16 +1330,19 @@ check_flags(SupFlags) when is_map(SupFlags) ->
 check_flags({Strategy, MaxIntensity, Period}) ->
     check_flags(#{strategy => Strategy,
 		  intensity => MaxIntensity,
-		  period => Period});
+		  period => Period,
+		  auto_shutdown => never});
 check_flags(What) ->
     throw({invalid_type, What}).
 
 do_check_flags(#{strategy := Strategy,
 		 intensity := MaxIntensity,
-		 period := Period} = Flags) ->
+		 period := Period,
+		 auto_shutdown := AutoShutdown} = Flags) ->
     validStrategy(Strategy),
     validIntensity(MaxIntensity),
     validPeriod(Period),
+    validAutoShutdown(AutoShutdown),
     Flags.
 
 validStrategy(simple_one_for_one) -> true;
@@ -1264,45 +1359,59 @@ validPeriod(Period) when is_integer(Period),
                          Period > 0 -> true;
 validPeriod(What)                   -> throw({invalid_period, What}).
 
+validAutoShutdown(never)           -> true;
+validAutoShutdown(any_significant) -> true;
+validAutoShutdown(all_significant) -> true;
+validAutoShutdown(What)            -> throw({invalid_auto_shutdown, What}).
+
+
 supname(self, Mod) -> {self(), Mod};
 supname(N, _)      -> N.
 
 %%% ------------------------------------------------------
 %%% Check that the children start specification is valid.
 %%% Input: [child_spec()]
+%%%        auto_shutdown()
 %%% Returns: {ok, [child_rec()]} | Error
 %%% ------------------------------------------------------
 
-check_startspec(Children) -> check_startspec(Children, [], #{}).
+check_startspec(Children, AutoShutdown) ->
+    check_startspec(Children, [], #{}, AutoShutdown).
 
-check_startspec([ChildSpec|T], Ids, Db) ->
-    case check_childspec(ChildSpec) of
+check_startspec([ChildSpec|T], Ids, Db, AutoShutdown) ->
+    case check_childspec(ChildSpec, AutoShutdown) of
 	{ok, #child{id=Id}=Child} ->
 	    case maps:is_key(Id, Db) of
 		%% The error message duplicate_child_name is kept for
 		%% backwards compatibility, although
 		%% duplicate_child_id would be more correct.
 		true -> {duplicate_child_name, Id};
-		false -> check_startspec(T, [Id | Ids], Db#{Id=>Child})
+		false -> check_startspec(T, [Id | Ids], Db#{Id=>Child},
+					 AutoShutdown)
 	    end;
 	Error -> Error
     end;
-check_startspec([], Ids, Db) ->
+check_startspec([], Ids, Db, _AutoShutdown) ->
     {ok, {lists:reverse(Ids),Db}}.
 
-check_childspec(ChildSpec) when is_map(ChildSpec) ->
-    catch do_check_childspec(maps:merge(?default_child_spec,ChildSpec));
-check_childspec({Id, Func, RestartType, Shutdown, ChildType, Mods}) ->
+check_childspec(ChildSpec, AutoShutdown) when is_map(ChildSpec) ->
+    catch do_check_childspec(maps:merge(?default_child_spec,ChildSpec),
+			     AutoShutdown);
+check_childspec({Id, Func, RestartType, Shutdown, ChildType, Mods},
+		AutoShutdown) ->
     check_childspec(#{id => Id,
 		      start => Func,
 		      restart => RestartType,
+		      significant => false,
 		      shutdown => Shutdown,
 		      type => ChildType,
-		      modules => Mods});
-check_childspec(X) -> {invalid_child_spec, X}.
+		      modules => Mods},
+		    AutoShutdown);
+check_childspec(X, _AutoShutdown) -> {invalid_child_spec, X}.
 
 do_check_childspec(#{restart := RestartType,
-		     type := ChildType} = ChildSpec)->
+		     type := ChildType} = ChildSpec,
+		   AutoShutdown)->
     Id = case ChildSpec of
 	       #{id := I} -> I;
 	       _ -> throw(missing_id)
@@ -1314,6 +1423,11 @@ do_check_childspec(#{restart := RestartType,
     validId(Id),
     validFunc(Func),
     validRestartType(RestartType),
+    Significant = case ChildSpec of
+		      #{significant := Signf} -> Signf;
+		      _ -> false
+                  end,
+    validSignificant(Significant, RestartType, AutoShutdown),
     validChildType(ChildType),
     Shutdown = case ChildSpec of
 		   #{shutdown := S} -> S;
@@ -1327,7 +1441,8 @@ do_check_childspec(#{restart := RestartType,
 	   end,
     validMods(Mods),
     {ok, #child{id = Id, mfargs = Func, restart_type = RestartType,
-		shutdown = Shutdown, child_type = ChildType, modules = Mods}}.
+		significant = Significant, shutdown = Shutdown,
+		child_type = ChildType, modules = Mods}}.
 
 validChildType(supervisor) -> true;
 validChildType(worker) -> true;
@@ -1345,8 +1460,18 @@ validRestartType(temporary)   -> true;
 validRestartType(transient)   -> true;
 validRestartType(RestartType) -> throw({invalid_restart_type, RestartType}).
 
+validSignificant(true, _RestartType, never) ->
+    throw({bad_combination, [{auto_shutdown, never}, {significant, true}]});
+validSignificant(true, permanent, _AutoShutdown) ->
+    throw({bad_combination, [{restart, permanent}, {significant, true}]});
+validSignificant(Significant, _RestartType, _AutoShutdown)
+  when is_boolean(Significant) ->
+    true;
+validSignificant(Significant, _RestartType, _AutoShutdown) ->
+    throw({invalid_significant, Significant}).
+
 validShutdown(Shutdown)
-  when is_integer(Shutdown), Shutdown > 0 -> true;
+  when is_integer(Shutdown), Shutdown >= 0 -> true;
 validShutdown(infinity)             -> true;
 validShutdown(brutal_kill)          -> true;
 validShutdown(Shutdown)             -> throw({invalid_shutdown, Shutdown}).
@@ -1365,12 +1490,14 @@ validMods(Mods) -> throw({invalid_modules, Mods}).
 child_to_spec(#child{id = Id,
 		    mfargs = Func,
 		    restart_type = RestartType,
+		    significant = Significant,
 		    shutdown = Shutdown,
 		    child_type = ChildType,
 		    modules = Mods}) ->
     #{id => Id,
       start => Func,
       restart => RestartType,
+      significant => Significant,
       shutdown => Shutdown,
       type => ChildType,
       modules => Mods}.
@@ -1389,7 +1516,7 @@ add_restart(State) ->
     P = State#state.period,
     R = State#state.restarts,
     Now = erlang:monotonic_time(1),
-    R1 = add_restart([Now|R], Now, P),
+    R1 = add_restart(R, Now, P),
     State1 = State#state{restarts = R1},
     case length(R1) of
 	CurI when CurI  =< I ->
@@ -1398,18 +1525,13 @@ add_restart(State) ->
 	    {terminate, State1}
     end.
 
-add_restart([R|Restarts], Now, Period) ->
-    case inPeriod(R, Now, Period) of
-	true ->
-	    [R|add_restart(Restarts, Now, Period)];
-	_ ->
-	    []
-    end;
-add_restart([], _, _) ->
-    [].
-
-inPeriod(Then, Now, Period) ->
-    Now =< Then + Period.
+add_restart(Restarts0, Now, Period) ->
+    Treshold = Now - Period,
+    Restarts1 = lists:takewhile(
+                  fun (R) -> R >= Treshold end,
+                  Restarts0
+                 ),
+    [Now | Restarts1].
 
 %%% ------------------------------------------------------
 %%% Error and progress reporting.
@@ -1419,6 +1541,7 @@ extract_child(Child) when is_list(Child#child.pid) ->
      {id, Child#child.id},
      {mfargs, Child#child.mfargs},
      {restart_type, Child#child.restart_type},
+     {significant, Child#child.significant},
      {shutdown, Child#child.shutdown},
      {child_type, Child#child.child_type}];
 extract_child(Child) ->
@@ -1426,6 +1549,7 @@ extract_child(Child) ->
      {id, Child#child.id},
      {mfargs, Child#child.mfargs},
      {restart_type, Child#child.restart_type},
+     {significant, Child#child.significant},
      {shutdown, Child#child.shutdown},
      {child_type, Child#child.child_type}].
 
@@ -1605,8 +1729,8 @@ format_status(_, [_PDict, State]) ->
 dyn_size(#state{dynamics = {_Kind,Db}}) ->
     map_size(Db).
 
-dyn_erase(Pid,#state{dynamics={_Kind,Db}}=State) ->
-    State#state{dynamics={maps,maps:remove(Pid,Db)}}.
+dyn_erase(Pid,#state{dynamics={Kind,Db}}=State) ->
+    State#state{dynamics={Kind,maps:remove(Pid,Db)}}.
 
 dyn_store(Pid,Args,#state{dynamics={Kind,Db}}=State) ->
     case Kind of
